# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import fixtures
import flask
import flask_restful
import functools
from oslo_policy import policy
from oslo_serialization import jsonutils
from testtools import matchers

from keystone.common import context
from keystone.common import json_home
from keystone.common import rbac_enforcer
import keystone.conf
from keystone import exception
from keystone.server.flask import common as flask_common
from keystone.server.flask.request_processing import json_body
from keystone.tests.unit import rest


CONF = keystone.conf.CONF


class _TestResourceWithCollectionInfo(flask_common.ResourceBase):
    collection_key = 'arguments'
    member_key = 'argument'
    __shared_state__ = {}
    _storage_dict = {}

    def __init__(self):
        super(_TestResourceWithCollectionInfo, self).__init__()
        # Share State, this is for "dummy" backend storage.
        self.__dict__ = self.__shared_state__

    @classmethod
    def _reset(cls):
        # Used after a test to ensure clean-state
        cls._storage_dict.clear()
        cls.__shared_state__.clear()

    def _list_arguments(self):
        return self.wrap_collection(list(self._storage_dict.values()))

    def get(self, argument_id=None):
        # List with no argument, get resource with id, used for HEAD as well.
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed')
        if argument_id is None:
            # List
            return self._list_arguments()
        else:
            # get resource with id
            try:
                return self.wrap_member(self._storage_dict[argument_id])
            except KeyError:
                raise exception.NotFound(target=argument_id)

    def post(self):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed')
        ref = flask.request.get_json(force=True)
        ref = self._assign_unique_id(ref)
        self._storage_dict[ref['id']] = ref
        return self.wrap_member(self._storage_dict[ref['id']]), 201

    def put(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed')
        try:
            self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        ref = flask.request.get_json(force=True)
        self._require_matching_id(ref)
        # Maintain the ref id
        ref['id'] = argument_id
        self._storage_dict[argument_id] = ref
        return '', 204

    def patch(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed')
        try:
            self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        ref = flask.request.get_json(force=True)
        self._require_matching_id(ref)
        self._storage_dict[argument_id].update(ref)
        return self.wrap_member(self._storage_dict[argument_id])

    def delete(self, argument_id):
        rbac_enforcer.enforcer.RBACEnforcer.enforce_call(
            action='example:allowed')
        try:
            del self._storage_dict[argument_id]
        except KeyError:
            raise exception.NotFound(target=argument_id)
        return '', 204


class _TestRestfulAPI(flask_common.APIBase):
    _name = 'test_api_base'
    _import_name = __name__
    resources = []
    resource_mapping = []

    def __init__(self, *args, **kwargs):
        self.resource_mapping = kwargs.pop('resource_mapping', [])
        self.resources = kwargs.pop('resources',
                                    [_TestResourceWithCollectionInfo])
        super(_TestRestfulAPI, self).__init__(*args, **kwargs)


class TestKeystoneFlaskCommon(rest.RestfulTestCase):

    _policy_rules = [
        policy.RuleDefault(
            name='example:allowed',
            check_str=''
        ),
        policy.RuleDefault(
            name='example:deny',
            check_str='false:false'
        )
    ]

    def setUp(self):
        super(TestKeystoneFlaskCommon, self).setUp()
        enf = rbac_enforcer.enforcer.RBACEnforcer()

        def register_rules(enf_obj):
            enf_obj.register_defaults(self._policy_rules)

        self.useFixture(fixtures.MockPatchObject(
            enf, 'register_rules', register_rules))
        self.useFixture(fixtures.MockPatchObject(
            rbac_enforcer.enforcer, '_POSSIBLE_TARGET_ACTIONS',
            {r.name for r in self._policy_rules}))

        enf._reset()
        self.addCleanup(enf._reset)
        self.addCleanup(
            _TestResourceWithCollectionInfo._reset)

    def _get_token(self):
        auth_json = {
            'auth': {
                'identity': {
                    'methods': ['password'],
                    'password': {
                        'user': {
                            'name': self.user_req_admin['name'],
                            'password': self.user_req_admin['password'],
                            'domain': {
                                'id': self.user_req_admin['domain_id']
                            }
                        }
                    }
                },
                'scope': {
                    'project': {
                        'id': self.project_service['id']
                    }
                }
            }
        }
        return self.test_client().post(
            '/v3/auth/tokens',
            json=auth_json,
            expected_status_code=201).headers['X-Subject-Token']

    def _setup_flask_restful_api(self, **options):

        self.restful_api_opts = options.copy()
        orig_value = _TestResourceWithCollectionInfo.api_prefix
        setattr(_TestResourceWithCollectionInfo,
                'api_prefix', options.get('api_url_prefix', ''))
        self.addCleanup(setattr, _TestResourceWithCollectionInfo, 'api_prefix',
                        orig_value)
        self.restful_api = _TestRestfulAPI(**options)
        self.public_app.app.register_blueprint(self.restful_api.blueprint)
        self.cleanup_instance('restful_api')
        self.cleanup_instance('restful_api_opts')

    def _make_requests(self):
        path_base = '/arguments'
        api_prefix = self.restful_api_opts.get('api_url_prefix', '')
        blueprint_prefix = self.restful_api._blueprint_url_prefix.rstrip('/')
        url = ''.join(
            [x for x in [blueprint_prefix, api_prefix, path_base] if x])
        headers = {'X-Auth-Token': self._get_token()}
        with self.test_client() as c:
            # GET LIST
            resp = c.get(url, headers=headers)
            self.assertEqual(
                _TestResourceWithCollectionInfo.wrap_collection(
                    []), resp.json)
            unknown_id = uuid.uuid4().hex

            # GET non-existent ref
            c.get('%s/%s' % (url, unknown_id), headers=headers,
                  expected_status_code=404)

            # HEAD non-existent ref
            c.head('%s/%s' % (url, unknown_id), headers=headers,
                   expected_status_code=404)

            # PUT non-existent ref
            c.put('%s/%s' % (url, unknown_id), json={}, headers=headers,
                  expected_status_code=404)

            # PATCH non-existent ref
            c.patch('%s/%s' % (url, unknown_id), json={}, headers=headers,
                    expected_status_code=404)

            # DELETE non-existent ref
            c.delete('%s/%s' % (url, unknown_id), headers=headers,
                     expected_status_code=404)

            # POST new ref
            new_argument_resource = {'testing': uuid.uuid4().hex}
            new_argument_resp = c.post(
                url,
                json=new_argument_resource,
                headers=headers).json['argument']

            # POST second new ref
            new_argument2_resource = {'testing': uuid.uuid4().hex}
            new_argument2_resp = c.post(
                url,
                json=new_argument2_resource,
                headers=headers).json['argument']

            # GET list
            get_list_resp = c.get(url, headers=headers).json
            self.assertIn(new_argument_resp,
                          get_list_resp['arguments'])
            self.assertIn(new_argument2_resp,
                          get_list_resp['arguments'])

            # GET first ref
            get_resp = c.get('%s/%s' % (url, new_argument_resp['id']),
                             headers=headers).json['argument']
            self.assertEqual(new_argument_resp, get_resp)

            # HEAD first ref
            head_resp = c.head(
                '%s/%s' % (url, new_argument_resp['id']),
                headers=headers).data
            # NOTE(morgan): For python3 compat, explicitly binary type
            self.assertEqual(head_resp, b'')

            # PUT update first ref
            replacement_argument = {'new_arg': True, 'id': uuid.uuid4().hex}
            c.put('%s/%s' % (url, new_argument_resp['id']), headers=headers,
                  json=replacement_argument, expected_status_code=400)
            replacement_argument.pop('id')
            c.put('%s/%s' % (url, new_argument_resp['id']),
                  headers=headers,
                  json=replacement_argument)
            put_resp = c.get('%s/%s' % (url, new_argument_resp['id']),
                             headers=headers).json['argument']
            self.assertNotIn(new_argument_resp['testing'],
                             put_resp)
            self.assertTrue(put_resp['new_arg'])

            # GET first ref (check for replacement)
            get_replacement_resp = c.get(
                '%s/%s' % (url, new_argument_resp['id']),
                headers=headers).json['argument']
            self.assertEqual(put_resp,
                             get_replacement_resp)

            # PATCH update first ref
            patch_ref = {'uuid': uuid.uuid4().hex}
            patch_resp = c.patch('%s/%s' % (url, new_argument_resp['id']),
                                 headers=headers,
                                 json=patch_ref).json['argument']
            self.assertTrue(patch_resp['new_arg'])
            self.assertEqual(patch_ref['uuid'], patch_resp['uuid'])

            # GET first ref (check for update)
            get_patched_ref_resp = c.get(
                '%s/%s' % (url, new_argument_resp['id']),
                headers=headers).json['argument']
            self.assertEqual(patch_resp,
                             get_patched_ref_resp)

            # DELETE first ref
            c.delete(
                '%s/%s' % (url, new_argument_resp['id']),
                headers=headers)
            # Check that it was in-fact deleted
            c.get(
                '%s/%s' % (url, new_argument_resp['id']),
                headers=headers, expected_status_code=404)

    def test_api_url_prefix(self):
        url_prefix = '/%s' % uuid.uuid4().hex
        self._setup_flask_restful_api(
            api_url_prefix=url_prefix)
        self._make_requests()

    def test_blueprint_url_prefix(self):
        url_prefix = '/%s' % uuid.uuid4().hex
        self._setup_flask_restful_api(
            blueprint_url_prefix=url_prefix)
        self._make_requests()

    def test_build_restful_api_no_prefix(self):
        self._setup_flask_restful_api()
        self._make_requests()

    def test_cannot_add_before_request_functions_twice(self):

        class TestAPIDuplicateBefore(_TestRestfulAPI):
            def __init__(self):
                super(TestAPIDuplicateBefore, self).__init__()
                self._register_before_request_functions()

        self.assertRaises(AssertionError, TestAPIDuplicateBefore)

    def test_cannot_add_after_request_functions_twice(self):

        class TestAPIDuplicateAfter(_TestRestfulAPI):
            def __init__(self):
                super(TestAPIDuplicateAfter, self).__init__()
                self._register_after_request_functions()

        self.assertRaises(AssertionError, TestAPIDuplicateAfter)

    def test_after_request_functions_must_be_added(self):

        class TestAPINoAfter(_TestRestfulAPI):
            def _register_after_request_functions(self, functions=None):
                pass

        self.assertRaises(AssertionError, TestAPINoAfter)

    def test_before_request_functions_must_be_added(self):

        class TestAPINoBefore(_TestRestfulAPI):
            def _register_before_request_functions(self, functions=None):
                pass

        self.assertRaises(AssertionError, TestAPINoBefore)

    def test_before_request_functions(self):
        # Test additional "before" request functions fire.
        attr = uuid.uuid4().hex

        def do_something():
            setattr(flask.g, attr, True)

        class TestAPI(_TestRestfulAPI):
            def _register_before_request_functions(self, functions=None):
                functions = functions or []
                functions.append(do_something)
                super(TestAPI, self)._register_before_request_functions(
                    functions)

        api = TestAPI(resources=[_TestResourceWithCollectionInfo])
        self.public_app.app.register_blueprint(api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            c.get('/v3/arguments', headers={'X-Auth-Token': token})
            self.assertTrue(getattr(flask.g, attr, False))

    def test_after_request_functions(self):
        # Test additional "after" request functions fire. In this case, we
        # alter the response code to 420
        attr = uuid.uuid4().hex

        def do_something(resp):
            setattr(flask.g, attr, True)
            resp.status_code = 420
            return resp

        class TestAPI(_TestRestfulAPI):
            def _register_after_request_functions(self, functions=None):
                functions = functions or []
                functions.append(do_something)
                super(TestAPI, self)._register_after_request_functions(
                    functions)

        api = TestAPI(resources=[_TestResourceWithCollectionInfo])
        self.public_app.app.register_blueprint(api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            c.get('/v3/arguments', headers={'X-Auth-Token': token},
                  expected_status_code=420)

    def test_construct_resource_map(self):
        resource_name = 'arguments'
        param_relation = json_home.build_v3_parameter_relation(
            'argument_id')
        alt_rel_func = functools.partial(
            json_home.build_v3_extension_resource_relation,
            extension_name='extension', extension_version='1.0')
        url = '/v3/arguments/<string:argument_id>'
        old_url = [dict(
            url='/v3/old_arguments/<string:argument_id>',
            json_home=flask_common.construct_json_home_data(
                rel='arguments',
                resource_relation_func=alt_rel_func)
        )]

        mapping = flask_common.construct_resource_map(
            resource=_TestResourceWithCollectionInfo,
            url=url,
            resource_kwargs={},
            alternate_urls=old_url,
            rel=resource_name,
            status=json_home.Status.EXPERIMENTAL,
            path_vars={'argument_id': param_relation},
            resource_relation_func=json_home.build_v3_resource_relation)
        self.assertEqual(_TestResourceWithCollectionInfo,
                         mapping.resource)
        self.assertEqual(url, mapping.url)
        self.assertEqual(json_home.build_v3_resource_relation(resource_name),
                         mapping.json_home_data.rel)
        self.assertEqual(json_home.Status.EXPERIMENTAL,
                         mapping.json_home_data.status)
        self.assertEqual({'argument_id': param_relation},
                         mapping.json_home_data.path_vars)
        # Check the alternate URL data is populated sanely
        self.assertEqual(1, len(mapping.alternate_urls))
        alt_url_data = mapping.alternate_urls[0]
        self.assertEqual(old_url[0]['url'], alt_url_data['url'])
        self.assertEqual(old_url[0]['json_home'], alt_url_data['json_home'])

    def test_instantiate_and_register_to_app(self):
        # Test that automatic instantiation and registration to app works.
        self.restful_api_opts = {}
        self.restful_api = _TestRestfulAPI.instantiate_and_register_to_app(
            self.public_app.app)
        self.cleanup_instance('restful_api_opts')
        self.cleanup_instance('restful_api')
        self._make_requests()

    def test_unenforced_api_decorator(self):
        # Test unenforced decorator works as expected

        class MappedResource(flask_restful.Resource):
            @flask_common.unenforced_api
            def post(self):
                post_body = flask.request.get_json()
                return {'post_body': post_body}, 201

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation)

        restful_api = _TestRestfulAPI(resource_mapping=[resource_map],
                                      resources=[])
        self.public_app.app.register_blueprint(restful_api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            body = {'test_value': uuid.uuid4().hex}
            # Works with token
            resp = c.post('/v3/test_api', json=body,
                          headers={'X-Auth-Token': token})
            self.assertEqual(body, resp.json['post_body'])
            # Works without token
            resp = c.post('/v3/test_api', json=body)
            self.assertEqual(body, resp.json['post_body'])

    def test_HTTP_OPTIONS_is_unenforced(self):
        # Standup a test mapped resource and call OPTIONS on it. This will
        # return a header "Allow" with the valid methods, in this case
        # OPTIONS and POST. Ensure that the response otherwise conforms
        # as expected.

        class MappedResource(flask_restful.Resource):
            def post(self):
                # we don't actually use this or call it.
                pass

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation)

        restful_api = _TestRestfulAPI(resource_mapping=[resource_map],
                                      resources=[])
        self.public_app.app.register_blueprint(restful_api.blueprint)
        with self.test_client() as c:
            r = c.options('/v3/test_api')
            # make sure we split the data and left/right strip off whitespace
            # The use of a SET here is to ensure the exact values are in place
            # even if hash-seeds change order. `r.data` will be an empty
            # byte-string. `Content-Length` will be 0.
            self.assertEqual(
                set(['OPTIONS', 'POST']),
                set([v.lstrip().rstrip()
                     for v in r.headers['Allow'].split(',')]))
            self.assertEqual(r.headers['Content-Length'], '0')
            self.assertEqual(r.data, b'')

    def test_mapped_resource_routes(self):
        # Test non-standard URL routes ("mapped") function as expected

        class MappedResource(flask_restful.Resource):
            def post(self):
                rbac_enforcer.enforcer.RBACEnforcer().enforce_call(
                    action='example:allowed')
                post_body = flask.request.get_json()
                return {'post_body': post_body}, 201

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation)

        restful_api = _TestRestfulAPI(resource_mapping=[resource_map],
                                      resources=[])
        self.public_app.app.register_blueprint(restful_api.blueprint)
        token = self._get_token()
        with self.test_client() as c:
            body = {'test_value': uuid.uuid4().hex}
            resp = c.post('/v3/test_api', json=body,
                          headers={'X-Auth-Token': token})
            self.assertEqual(body, resp.json['post_body'])

    def test_correct_json_home_document(self):
        class MappedResource(flask_restful.Resource):
            def post(self):
                rbac_enforcer.enforcer.RBACEnforcer().enforce_call(
                    action='example:allowed')
                post_body = flask.request.get_json()
                return {'post_body': post_body}

        # NOTE(morgan): totally fabricated json_home data based upon our TEST
        # restful_apis.
        json_home_data = {
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/argument': {
                'href-template': '/v3/arguments/{argument_id}',
                'href-vars': {
                    'argument_id': 'https://docs.openstack.org/api/'
                                   'openstack-identity/3/param/argument_id'
                }
            },
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/arguments': {
                'href': '/v3/arguments'
            },
            'https://docs.openstack.org/api/openstack-identity/3/'
            'rel/test': {
                'href': '/v3/test_api'
            },
        }

        resource_map = flask_common.construct_resource_map(
            resource=MappedResource,
            url='test_api',
            alternate_urls=[],
            resource_kwargs={},
            rel='test',
            status=json_home.Status.STABLE,
            path_vars=None,
            resource_relation_func=json_home.build_v3_resource_relation)

        restful_api = _TestRestfulAPI(resource_mapping=[resource_map])
        self.public_app.app.register_blueprint(restful_api.blueprint)

        with self.test_client() as c:
            headers = {'Accept': 'application/json-home'}
            resp = c.get('/', headers=headers)
            resp_data = jsonutils.loads(resp.data)
            for rel in json_home_data:
                self.assertThat(resp_data['resources'][rel],
                                matchers.Equals(json_home_data[rel]))

    def test_normalize_domain_id_extracts_domain_id_if_needed(self):
        self._setup_flask_restful_api()
        blueprint_prefix = self.restful_api._blueprint_url_prefix.rstrip('/')
        url = ''.join([blueprint_prefix, '/arguments'])
        headers = {'X-Auth-Token': self._get_token()}
        ref_with_domain_id = {'domain_id': uuid.uuid4().hex}
        ref_without_domain_id = {}
        with self.test_client() as c:
            # Make a dummy request.. ANY request is fine to push the whole
            # context stack.
            c.get('%s/%s' % (url, uuid.uuid4().hex), headers=headers,
                  expected_status_code=404)

            oslo_context = flask.request.environ[
                context.REQUEST_CONTEXT_ENV]

            # Normal Project Scope Form
            # ---------------------------
            # test that _normalize_domain_id does something sane
            domain_id = ref_with_domain_id['domain_id']
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            # Ensure (deprecated) we add default domain if needed
            flask_common.ResourceBase._normalize_domain_id(
                ref_without_domain_id)
            self.assertEqual(
                CONF.identity.default_domain_id,
                ref_without_domain_id['domain_id'])
            ref_without_domain_id.clear()

            # Domain Scoped Form
            # --------------------
            # Just set oslo_context domain_id to a value. This is how we
            # communicate domain scope. No need to explicitly
            # do a domain-scoped request, this is a synthetic text anyway
            oslo_context.domain_id = uuid.uuid4().hex
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            flask_common.ResourceBase._normalize_domain_id(
                ref_without_domain_id)
            self.assertEqual(oslo_context.domain_id,
                             ref_without_domain_id['domain_id'])
            ref_without_domain_id.clear()

            # "Admin" Token form
            # -------------------
            # Explicitly set "is_admin" to true, no new request is needed
            # as we simply check "is_admin" value everywhere
            oslo_context.is_admin = True
            oslo_context.domain_id = None
            # Ensure we don't change the domain if it is specified
            flask_common.ResourceBase._normalize_domain_id(ref_with_domain_id)
            self.assertEqual(domain_id, ref_with_domain_id['domain_id'])
            # Ensure we raise an appropriate exception with the inferred
            # domain_id
            self.assertRaises(exception.ValidationError,
                              flask_common.ResourceBase._normalize_domain_id,
                              ref=ref_without_domain_id)

    def test_api_prefix_self_referential_link_substitution(self):

        view_arg = uuid.uuid4().hex

        class TestResource(flask_common.ResourceBase):
            api_prefix = '/<string:test_value>/nothing'

        # use a dummy request context, no enforcement is happening
        # therefore we don't need the heavy lifting of a full request
        # run.
        with self.test_request_context(
                path='/%s/nothing/values' % view_arg,
                base_url='https://localhost/'):
            # explicitly set the view_args, this is a special case
            # for a synthetic test case, usually one would rely on
            # a full request stack to set these.
            flask.request.view_args = {'test_value': view_arg}

            # create dummy ref
            ref = {'id': uuid.uuid4().hex}

            # add the self referential link
            TestResource._add_self_referential_link(
                ref, collection_name='values')

            # Check that the link in fact starts with what we expect
            # including the explicit view arg.
            self.assertTrue(ref['links']['self'].startswith(
                'https://localhost/v3/%s' % view_arg)
            )

    def test_json_body_before_req_func_valid_json(self):
        with self.test_request_context(
                headers={'Content-Type': 'application/json'},
                data='{"key": "value"}'):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

    def test_json_body_before_req_func_invalid_json(self):
        with self.test_request_context(
                headers={'Content-Type': 'application/json'},
                data='invalid JSON'):
            # keystone.exception.ValidationError should be raised
            self.assertRaises(exception.ValidationError,
                              json_body.json_body_before_request)

    def test_json_body_before_req_func_no_content_type(self):
        # Unset
        with self.test_request_context(data='{"key": "value"}'):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

        # Explicitly set to ''
        with self.test_request_context(
                headers={'Content-Type': ''}, data='{"key": "value"}'):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

    def test_json_body_before_req_func_unrecognized_content_type(self):
        with self.test_request_context(
                headers={'Content-Type': 'unrecognized/content-type'},
                data='{"key": "value"'):
            # keystone.exception.ValidationError should be raised
            self.assertRaises(exception.ValidationError,
                              json_body.json_body_before_request)

    def test_json_body_before_req_func_unrecognized_conten_type_no_body(self):
        with self.test_request_context(
                headers={'Content-Type': 'unrecognized/content-type'}):
            # No exception should be raised, everything is happy.
            json_body.json_body_before_request()

    def test_resource_collection_key_raises_exception_if_unset(self):
        class TestResource(flask_common.ResourceBase):
            """A Test Resource."""

        class TestResourceWithKey(flask_common.ResourceBase):
            collection_key = uuid.uuid4().hex

        r = TestResource()
        self.assertRaises(ValueError, getattr, r, 'collection_key')

        r = TestResourceWithKey()
        self.assertEqual(
            TestResourceWithKey.collection_key, r.collection_key)

    def test_resource_member_key_raises_exception_if_unset(self):
        class TestResource(flask_common.ResourceBase):
            """A Test Resource."""

        class TestResourceWithKey(flask_common.ResourceBase):
            member_key = uuid.uuid4().hex

        r = TestResource()
        self.assertRaises(ValueError, getattr, r, 'member_key')

        r = TestResourceWithKey()
        self.assertEqual(
            TestResourceWithKey.member_key, r.member_key)


class TestKeystoneFlaskUnrouted404(rest.RestfulTestCase):
    def setUp(self):
        super(TestKeystoneFlaskUnrouted404, self).setUp()
        # unregister the 404 handler we explicitly set in loadapp. This
        # makes the 404 error fallback to a standard werkzeug handling.
        self.public_app.app.error_handler_spec[None].pop(404)

    def test_unrouted_path_is_not_jsonified_404(self):
        with self.test_client() as c:
            path = '/{unrouted_path}'.format(unrouted_path=uuid.uuid4())
            resp = c.get(path, expected_status_code=404)
            # Make sure we're emitting a html error
            self.assertEqual('text/html', resp.headers['Content-Type'])
            # Ensure the more generic flask/werkzeug 404 response is emitted
            self.assertTrue(b'404 Not Found' in resp.data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

deprecated_list_endpoint_groups = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoint_groups',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_get_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_list_projects_associated_with_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_projects_associated_with_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_list_endpoints_associated_with_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoints_associated_with_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_get_endpoint_group_in_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_endpoint_group_in_project',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_list_endpoint_groups_for_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_endpoint_groups_for_project',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_create_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_update_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_delete_endpoint_group = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_endpoint_group',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_add_endpoint_group_to_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'add_endpoint_group_to_project',
    check_str=base.RULE_ADMIN_REQUIRED,
)

deprecated_remove_endpoint_group_from_project = policy.DeprecatedRule(
    name=base.IDENTITY % 'remove_endpoint_group_from_project',
    check_str=base.RULE_ADMIN_REQUIRED,
)


DEPRECATED_REASON = (
    "The endpoint groups API is now aware of system scope and default roles."
)


group_endpoint_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_endpoint_group',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Create endpoint group.',
        operations=[{'path': '/v3/OS-EP-FILTER/endpoint_groups',
                     'method': 'POST'}],
        deprecated_rule=deprecated_create_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoint_groups',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List endpoint groups.',
        operations=[{'path': '/v3/OS-EP-FILTER/endpoint_groups',
                     'method': 'GET'}],
        deprecated_rule=deprecated_list_endpoint_groups,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_endpoint_group',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='Get endpoint group.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}'),
                     'method': 'GET'},
                    {'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}'),
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_get_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_endpoint_group',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Update endpoint group.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}'),
                     'method': 'PATCH'}],
        deprecated_rule=deprecated_update_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_endpoint_group',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Delete endpoint group.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}'),
                     'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_projects_associated_with_endpoint_group',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description=('List all projects associated with a specific endpoint '
                     'group.'),
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/projects'),
                     'method': 'GET'}],
        deprecated_rule=deprecated_list_projects_associated_with_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoints_associated_with_endpoint_group',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List all endpoints associated with an endpoint group.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/endpoints'),
                     'method': 'GET'}],
        deprecated_rule=deprecated_list_endpoints_associated_with_endpoint_group,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_endpoint_group_in_project',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description=('Check if an endpoint group is associated with a '
                     'project.'),
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/projects/{project_id}'),
                     'method': 'GET'},
                    {'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/projects/{project_id}'),
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_get_endpoint_group_in_project,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_endpoint_groups_for_project',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List endpoint groups associated with a specific project.',
        operations=[{'path': ('/v3/OS-EP-FILTER/projects/{project_id}/'
                              'endpoint_groups'),
                     'method': 'GET'}],
        deprecated_rule=deprecated_list_endpoint_groups_for_project,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'add_endpoint_group_to_project',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Allow a project to access an endpoint group.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/projects/{project_id}'),
                     'method': 'PUT'}],
        deprecated_rule=deprecated_add_endpoint_group_to_project,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'remove_endpoint_group_from_project',
        check_str=base.SYSTEM_ADMIN,
        scope_types=['system'],
        description='Remove endpoint group from project.',
        operations=[{'path': ('/v3/OS-EP-FILTER/endpoint_groups/'
                              '{endpoint_group_id}/projects/{project_id}'),
                     'method': 'DELETE'}],
        deprecated_rule=deprecated_remove_endpoint_group_from_project,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN)
]


def list_rules():
    return group_endpoint_policies



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from six.moves import http_client

from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _UserAccessRuleTests(object):
    """Test cases for anyone that has a valid user token."""

    def test_user_can_get_their_access_rules(self):
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (self.user_id, app_cred['access_rules'][0]['id'])
            c.get(path, headers=self.headers)

    def test_user_can_list_their_access_rules(self):
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': uuid.uuid4().hex,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        with self.test_client() as c:
            r = c.get('/v3/users/%s/access_rules' % self.user_id, headers=self.headers)
            self.assertEqual(len(r.json['access_rules']), 1)

    def test_user_can_delete_their_access_rules(self):
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': self.user_id,
            'project_id': self.project_id,
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        PROVIDERS.application_credential_api.delete_application_credential(app_cred['id'])
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (self.user_id, access_rule_id)
            c.delete(path, headers=self.headers)


class _ProjectUsersTests(object):
    """Users who have project role authorization observe the same behavior."""

    def test_user_cannot_get_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.get(
                path, headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

    def test_user_cannot_get_own_non_existent_access_rule_not_found(self):
        with self.test_client() as c:
            c.get(
                '/v3/users/%s/access_rules/%s' % (
                    self.user_id, uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.NOT_FOUND
            )

    def test_user_cannot_get_non_existent_access_rule_other_user_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.get(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

    def test_user_cannot_list_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': uuid.uuid4().hex,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)

        with self.test_client() as c:
            path = '/v3/users/%s/access_rules' % user['id']
            c.get(path, headers=self.headers,
                  expected_status_code=http_client.FORBIDDEN)

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        PROVIDERS.application_credential_api.delete_application_credential(app_cred['id'])
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.delete(
                path, headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existent_access_rule_other_user_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )


class _SystemUserAccessRuleTests(object):
    """Tests that are common across all system users."""

    def test_user_can_list_access_rules_for_other_users(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )

        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': uuid.uuid4().hex,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)

        with self.test_client() as c:
            r = c.get('/v3/users/%s/access_rules' % user['id'],
                      headers=self.headers)
            self.assertEqual(1, len(r.json['access_rules']))

    def test_user_cannot_get_non_existent_access_rule_not_found(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.get(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.NOT_FOUND
            )


class SystemReaderTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _SystemUserAccessRuleTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_reader
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_reader['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        PROVIDERS.application_credential_api.delete_application_credential(app_cred['id'])
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.delete(
                path, headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existent_access_rule_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )


class SystemMemberTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _SystemUserAccessRuleTests):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            system_member
        )['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id, password=system_member['password'],
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_cannot_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )

        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        PROVIDERS.application_credential_api.delete_application_credential(app_cred['id'])
        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.delete(
                path, headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.delete(
                path, headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )

    def test_user_cannot_delete_non_existent_access_rule_forbidden(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.FORBIDDEN
            )


class SystemAdminTests(base_classes.TestCaseWithBootstrap,
                       common_auth.AuthTestMixin,
                       _SystemUserAccessRuleTests):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_delete_access_rules_for_others(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=user['id'],
            project_id=project['id']
        )
        access_rule_id = uuid.uuid4().hex
        app_cred = {
            'id': uuid.uuid4().hex,
            'name': uuid.uuid4().hex,
            'user_id': user['id'],
            'project_id': project['id'],
            'secret': uuid.uuid4().hex,
            'access_rules': [{
                'id': access_rule_id,
                'service': uuid.uuid4().hex,
                'path': uuid.uuid4().hex,
                'method': uuid.uuid4().hex[16:]
            }]
        }
        PROVIDERS.application_credential_api.create_application_credential(app_cred)
        PROVIDERS.application_credential_api.delete_application_credential(app_cred['id'])

        with self.test_client() as c:
            path = '/v3/users/%s/access_rules/%s' % (user['id'], access_rule_id)
            c.delete(path, headers=self.headers)

    def test_user_cannot_delete_non_existent_access_rule_not_found(self):
        user = unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        user = PROVIDERS.identity_api.create_user(user)
        with self.test_client() as c:
            c.delete(
                '/v3/users/%s/access_rules/%s' % (
                    user['id'], uuid.uuid4().hex),
                headers=self.headers,
                expected_status_code=http_client.NOT_FOUND
            )


class ProjectReaderTests(base_classes.TestCaseWithBootstrap,
                         common_auth.AuthTestMixin,
                         _UserAccessRuleTests,
                         _ProjectUsersTests):

    def setUp(self):
        super(ProjectReaderTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_reader
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(base_classes.TestCaseWithBootstrap,
                         common_auth.AuthTestMixin,
                         _UserAccessRuleTests,
                         _ProjectUsersTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(
            project_member
        )['id']
        project = unit.new_project_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.project_id = PROVIDERS.resource_api.create_project(
            project['id'], project
        )['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id, user_id=self.user_id,
            project_id=self.project_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(base_classes.TestCaseWithBootstrap,
                        common_auth.AuthTestMixin,
                        _UserAccessRuleTests,
                        _ProjectUsersTests):

    def setUp(self):
        super(ProjectAdminTests, self).setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        # Reuse the system administrator account created during
        # ``keystone-manage bootstrap``
        self.user_id = self.bootstrapper.admin_user_id
        self.project_id = self.bootstrapper.project_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.project_id
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

RULE_TRUSTOR = 'user_id:%(target.trust.trustor_user_id)s'
RULE_TRUSTEE = 'user_id:%(target.trust.trustee_user_id)s'
SYSTEM_READER_OR_TRUSTOR_OR_TRUSTEE = (
    base.SYSTEM_READER + ' or ' + RULE_TRUSTOR + ' or ' + RULE_TRUSTEE
)
SYSTEM_READER_OR_TRUSTOR = base.SYSTEM_READER + ' or ' + RULE_TRUSTOR
SYSTEM_READER_OR_TRUSTEE = base.SYSTEM_READER + ' or ' + RULE_TRUSTEE
SYSTEM_ADMIN_OR_TRUSTOR = base.SYSTEM_ADMIN + ' or ' + RULE_TRUSTOR

deprecated_list_trusts = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_trusts',
    check_str=base.RULE_ADMIN_REQUIRED
)
deprecated_list_roles_for_trust = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_roles_for_trust',
    check_str=RULE_TRUSTOR + ' or ' + RULE_TRUSTEE
)
deprecated_get_role_for_trust = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_role_for_trust',
    check_str=RULE_TRUSTOR + ' or ' + RULE_TRUSTEE
)
deprecated_delete_trust = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_trust',
    check_str=RULE_TRUSTOR
)
deprecated_get_trust = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_trust',
    check_str=RULE_TRUSTOR + ' or ' + RULE_TRUSTEE
)

DEPRECATED_REASON = (
    "The trust API is now aware of system scope and default roles."
)

trust_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_trust',
        check_str=base.RULE_TRUST_OWNER,
        # FIXME(lbragstad): Trusts have the ability to optionally include a
        # project, but until trusts deal with system scope it's not really
        # useful. For now, this should be a project only operation.
        scope_types=['project'],
        description='Create trust.',
        operations=[{'path': '/v3/OS-TRUST/trusts',
                     'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_trusts',
        check_str=base.SYSTEM_READER,
        scope_types=['system'],
        description='List trusts.',
        operations=[{'path': '/v3/OS-TRUST/trusts',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts',
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_list_trusts,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_trusts_for_trustor',
        check_str=SYSTEM_READER_OR_TRUSTOR,
        scope_types=['system', 'project'],
        description='List trusts for trustor.',
        operations=[{'path': '/v3/OS-TRUST/trusts?trustor_user_id={trustor_user_id}',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts?trustor_user_id={trustor_user_id}',
                     'method': 'HEAD'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_trusts_for_trustee',
        check_str=SYSTEM_READER_OR_TRUSTEE,
        scope_types=['system', 'project'],
        description='List trusts for trustee.',
        operations=[{'path': '/v3/OS-TRUST/trusts?trustee_user_id={trustee_user_id}',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts?trustee_user_id={trustee_user_id}',
                     'method': 'HEAD'}]),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_roles_for_trust',
        check_str=SYSTEM_READER_OR_TRUSTOR_OR_TRUSTEE,
        scope_types=['system', 'project'],
        description='List roles delegated by a trust.',
        operations=[{'path': '/v3/OS-TRUST/trusts/{trust_id}/roles',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts/{trust_id}/roles',
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_list_roles_for_trust,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_role_for_trust',
        check_str=SYSTEM_READER_OR_TRUSTOR_OR_TRUSTEE,
        scope_types=['system', 'project'],
        description='Check if trust delegates a particular role.',
        operations=[{'path': '/v3/OS-TRUST/trusts/{trust_id}/roles/{role_id}',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts/{trust_id}/roles/{role_id}',
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_get_role_for_trust,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_trust',
        check_str=SYSTEM_ADMIN_OR_TRUSTOR,
        scope_types=['system', 'project'],
        description='Revoke trust.',
        operations=[{'path': '/v3/OS-TRUST/trusts/{trust_id}',
                     'method': 'DELETE'}],
        deprecated_rule=deprecated_delete_trust,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_trust',
        check_str=SYSTEM_READER_OR_TRUSTOR_OR_TRUSTEE,
        scope_types=['system', 'project'],
        description='Get trust.',
        operations=[{'path': '/v3/OS-TRUST/trusts/{trust_id}',
                     'method': 'GET'},
                    {'path': '/v3/OS-TRUST/trusts/{trust_id}',
                     'method': 'HEAD'}],
        deprecated_rule=deprecated_get_trust,
        deprecated_reason=DEPRECATED_REASON,
        deprecated_since=versionutils.deprecated.TRAIN)
]


def list_rules():
    return trust_policies



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sql

from keystone.common.sql import upgrades


def upgrade(migrate_engine):
    try:
        extension_version = upgrades.get_db_version(
            extension='oauth1',
            engine=migrate_engine)
    except Exception:
        extension_version = 0

    # This migration corresponds to oauth extension migration 5. Only
    # update if it has not been run.
    if extension_version >= 5:
        return

    # Upgrade operations go here. Don't create your own engine; bind
    # migrate_engine to your metadata
    meta = sql.MetaData()
    meta.bind = migrate_engine

    consumer_table = sql.Table(
        'consumer',
        meta,
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('description', sql.String(64), nullable=True),
        sql.Column('secret', sql.String(64), nullable=False),
        sql.Column('extra', sql.Text(), nullable=False))
    consumer_table.create(migrate_engine, checkfirst=True)

    request_token_table = sql.Table(
        'request_token',
        meta,
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('request_secret', sql.String(64), nullable=False),
        sql.Column('verifier', sql.String(64), nullable=True),
        sql.Column('authorizing_user_id', sql.String(64), nullable=True),
        sql.Column('requested_project_id', sql.String(64), nullable=False),
        sql.Column('role_ids', sql.Text(), nullable=True),
        sql.Column('consumer_id', sql.String(64),
                   sql.ForeignKey('consumer.id'),
                   nullable=False, index=True),
        sql.Column('expires_at', sql.String(64), nullable=True))
    request_token_table.create(migrate_engine, checkfirst=True)

    access_token_table = sql.Table(
        'access_token',
        meta,
        sql.Column('id', sql.String(64), primary_key=True, nullable=False),
        sql.Column('access_secret', sql.String(64), nullable=False),
        sql.Column('authorizing_user_id', sql.String(64),
                   nullable=False, index=True),
        sql.Column('project_id', sql.String(64), nullable=False),
        sql.Column('role_ids', sql.Text(), nullable=False),
        sql.Column('consumer_id', sql.String(64),
                   sql.ForeignKey('consumer.id'),
                   nullable=False, index=True),
        sql.Column('expires_at', sql.String(64), nullable=True))
    access_token_table.create(migrate_engine, checkfirst=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import constants
from keystone.conf import utils


methods = cfg.ListOpt(
    'methods',
    default=constants._DEFAULT_AUTH_METHODS,
    help=utils.fmt("""
Allowed authentication methods. Note: You should disable the `external` auth
method if you are currently using federation. External auth and federation
both use the REMOTE_USER variable. Since both the mapped and external plugin
are being invoked to validate attributes in the request environment, it can
cause conflicts.
"""))

password = cfg.StrOpt(  # nosec : This is the name of the plugin, not
    'password',         # a password that needs to be protected.
    help=utils.fmt("""
Entry point for the password auth plugin module in the `keystone.auth.password`
namespace. You do not need to set this unless you are overriding keystone's own
password authentication plugin.
"""))

token = cfg.StrOpt(
    'token',
    help=utils.fmt("""
Entry point for the token auth plugin module in the `keystone.auth.token`
namespace. You do not need to set this unless you are overriding keystone's own
token authentication plugin.
"""))

# deals with REMOTE_USER authentication
external = cfg.StrOpt(
    'external',
    help=utils.fmt("""
Entry point for the external (`REMOTE_USER`) auth plugin module in the
`keystone.auth.external` namespace. Supplied drivers are `DefaultDomain` and
`Domain`. The default driver is `DefaultDomain`, which assumes that all users
identified by the username specified to keystone in the `REMOTE_USER` variable
exist within the context of the default domain. The `Domain` option expects an
additional environment variable be presented to keystone, `REMOTE_DOMAIN`,
containing the domain name of the `REMOTE_USER` (if `REMOTE_DOMAIN` is not set,
then the default domain will be used instead). You do not need to set this
unless you are taking advantage of "external authentication", where the
application server (such as Apache) is handling authentication instead of
keystone.
"""))

oauth1 = cfg.StrOpt(
    'oauth1',
    help=utils.fmt("""
Entry point for the OAuth 1.0a auth plugin module in the `keystone.auth.oauth1`
namespace. You do not need to set this unless you are overriding keystone's own
`oauth1` authentication plugin.
"""))

mapped = cfg.StrOpt(
    'mapped',
    help=utils.fmt("""
Entry point for the mapped auth plugin module in the `keystone.auth.mapped`
namespace. You do not need to set this unless you are overriding keystone's own
`mapped` authentication plugin.
"""))

application_credential = cfg.StrOpt(
    'application_credential',
    help=utils.fmt("""
Entry point for the application_credential auth plugin module in the
`keystone.auth.application_credential` namespace. You do not need to set this
unless you are overriding keystone's own `application_credential`
authentication plugin.
"""))


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    methods,
    password,
    token,
    external,
    oauth1,
    mapped,
    application_credential,
]


def _register_auth_plugin_opt(conf, option):
    conf.register_opt(option, group=GROUP_NAME)


def setup_authentication(conf=None):
    """Register non-default auth methods (used by extensions, etc)."""
    # register any non-default auth methods here (used by extensions, etc)
    if conf is None:
        conf = cfg.CONF
    for method_name in conf.auth.methods:
        if method_name not in constants._DEFAULT_AUTH_METHODS:
            option = cfg.StrOpt(method_name)
            _register_auth_plugin_opt(conf, option)


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)

    setup_authentication(conf)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Copyright 2018 Catalyst Cloud Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Receipt provider interface."""

import datetime

from oslo_log import log
from oslo_utils import timeutils
import six

from keystone.common import cache
from keystone.common import manager
from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.i18n import _
from keystone.models import receipt_model
from keystone import notifications


CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs

RECEIPTS_REGION = cache.create_region(name='receipts')
MEMOIZE_RECEIPTS = cache.get_memoization_decorator(
    group='receipt',
    region=RECEIPTS_REGION)


def default_expire_time():
    """Determine when a fresh receipt should expire.

    Expiration time varies based on configuration (see
    ``[receipt] expiration``).

    :returns: a naive UTC datetime.datetime object

    """
    expire_delta = datetime.timedelta(seconds=CONF.receipt.expiration)
    expires_at = timeutils.utcnow() + expire_delta
    return expires_at.replace(microsecond=0)


class Manager(manager.Manager):
    """Default pivot point for the receipt provider backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.receipt.provider'
    _provides_api = 'receipt_provider_api'

    def __init__(self):
        super(Manager, self).__init__(CONF.receipt.provider)
        self._register_callback_listeners()

    def _register_callback_listeners(self):
        callbacks = {
            notifications.ACTIONS.deleted: [
                ['OS-TRUST:trust', self._drop_receipt_cache],
                ['user', self._drop_receipt_cache],
                ['domain', self._drop_receipt_cache],
            ],
            notifications.ACTIONS.disabled: [
                ['user', self._drop_receipt_cache],
                ['domain', self._drop_receipt_cache],
                ['project', self._drop_receipt_cache],
            ],
            notifications.ACTIONS.internal: [
                [notifications.INVALIDATE_TOKEN_CACHE,
                    self._drop_receipt_cache],
            ]
        }

        for event, cb_info in callbacks.items():
            for resource_type, callback_fns in cb_info:
                notifications.register_event_callback(event, resource_type,
                                                      callback_fns)

    def _drop_receipt_cache(self, service, resource_type, operation, payload):
        """Invalidate the entire receipt cache.

        This is a handy private utility method that should be used when
        consuming notifications that signal invalidating the receipt cache.

        """
        if CONF.receipt.cache_on_issue:
            RECEIPTS_REGION.invalidate()

    def validate_receipt(self, receipt_id, window_seconds=0):
        if not receipt_id:
            raise exception.ReceiptNotFound(
                _('No receipt in the request'), receipt_id=receipt_id)

        try:
            receipt = self._validate_receipt(receipt_id)
            self._is_valid_receipt(receipt, window_seconds=window_seconds)
            return receipt
        except exception.Unauthorized as e:
            LOG.debug('Unable to validate receipt: %s', e)
            raise exception.ReceiptNotFound(receipt_id=receipt_id)

    @MEMOIZE_RECEIPTS
    def _validate_receipt(self, receipt_id):
        (user_id, methods, issued_at,
            expires_at) = self.driver.validate_receipt(receipt_id)

        receipt = receipt_model.ReceiptModel()
        receipt.user_id = user_id
        receipt.methods = methods
        receipt.expires_at = expires_at
        receipt.mint(receipt_id, issued_at)
        return receipt

    def _is_valid_receipt(self, receipt, window_seconds=0):
        """Verify the receipt is valid format and has not expired."""
        current_time = timeutils.normalize_time(timeutils.utcnow())

        try:
            expiry = timeutils.parse_isotime(receipt.expires_at)
            expiry = timeutils.normalize_time(expiry)

            # add a window in which you can fetch a receipt beyond expiry
            expiry += datetime.timedelta(seconds=window_seconds)

        except Exception:
            LOG.exception('Unexpected error or malformed receipt '
                          'determining receipt expiry: %s', receipt)
            raise exception.ReceiptNotFound(
                _('Failed to validate receipt'), receipt_id=receipt.id)

        if current_time < expiry:
            return None
        else:
            raise exception.ReceiptNotFound(
                _('Failed to validate receipt'), receipt_id=receipt.id)

    def issue_receipt(self, user_id, method_names, expires_at=None):

        receipt = receipt_model.ReceiptModel()
        receipt.user_id = user_id
        receipt.methods = method_names

        if isinstance(expires_at, datetime.datetime):
            receipt.expires_at = utils.isotime(expires_at, subsecond=True)
        if isinstance(expires_at, six.string_types):
            receipt.expires_at = expires_at
        elif not expires_at:
            receipt.expires_at = utils.isotime(
                default_expire_time(), subsecond=True
            )

        receipt_id, issued_at = self.driver.generate_id_and_issued_at(receipt)
        receipt.mint(receipt_id, issued_at)

        if CONF.receipt.cache_on_issue:
            self._validate_receipt.set(
                receipt, RECEIPTS_REGION, receipt_id)

        return receipt



# Copyright 2013 OpenStack Foundation
# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import migrate
from migrate import exceptions
from migrate.versioning import api as versioning_api
from oslo_db import exception as db_exception
from oslo_db.sqlalchemy import migration
import six
import sqlalchemy

from keystone.common import sql
from keystone import exception
from keystone.i18n import _


USE_TRIGGERS = True

LEGACY_REPO = 'migrate_repo'
EXPAND_REPO = 'expand_repo'
DATA_MIGRATION_REPO = 'data_migration_repo'
CONTRACT_REPO = 'contract_repo'


class Repository(object):
    def __init__(self, engine, repo_name):
        self.repo_name = repo_name

        self.repo_path = find_repo(self.repo_name)
        self.min_version = (
            get_init_version(abs_path=self.repo_path))
        self.schema_ = versioning_api.ControlledSchema.create(
            engine, self.repo_path, self.min_version)
        self.max_version = self.schema_.repository.version().version

    def upgrade(self, version=None, current_schema=None):
        version = version or self.max_version
        err = ''
        upgrade = True
        version = versioning_api._migrate_version(
            self.schema_, version, upgrade, err)
        validate_upgrade_order(self.repo_name, target_repo_version=version)
        if not current_schema:
            current_schema = self.schema_
        changeset = current_schema.changeset(version)
        for ver, change in changeset:
            self.schema_.runchange(ver, change, changeset.step)

        if self.schema_.version != version:
            raise Exception(
                'Actual version (%s) of %s does not equal expected '
                'version (%s)' % (
                    self.schema_.version, self.repo_name, version))

    @property
    def version(self):
        with sql.session_for_read() as session:
            return migration.db_version(
                session.get_bind(), self.repo_path, self.min_version)


#  Different RDBMSs use different schemes for naming the Foreign Key
#  Constraints.  SQLAlchemy does not yet attempt to determine the name
#  for the constraint, and instead attempts to deduce it from the column.
#  This fails on MySQL.
def get_constraints_names(table, column_name):
    fkeys = [fk.name for fk in table.constraints
             if (isinstance(fk, sqlalchemy.ForeignKeyConstraint) and
                 column_name in fk.columns)]
    return fkeys


#  remove_constraints and add_constraints both accept a list of dictionaries
#  that contain:
#  {'table': a sqlalchemy table. The constraint is added to dropped from
#           this table.
#  'fk_column': the name of a column on the above table,  The constraint
#               is added to or dropped from this column
#  'ref_column':a sqlalchemy column object.  This is the reference column
#               for the constraint.
def remove_constraints(constraints):
    for constraint_def in constraints:
        constraint_names = get_constraints_names(constraint_def['table'],
                                                 constraint_def['fk_column'])
        for constraint_name in constraint_names:
            migrate.ForeignKeyConstraint(
                columns=[getattr(constraint_def['table'].c,
                                 constraint_def['fk_column'])],
                refcolumns=[constraint_def['ref_column']],
                name=constraint_name).drop()


def add_constraints(constraints):
    for constraint_def in constraints:

        if constraint_def['table'].kwargs.get('mysql_engine') == 'MyISAM':
            # Don't try to create constraint when using MyISAM because it's
            # not supported.
            continue

        ref_col = constraint_def['ref_column']
        ref_engine = ref_col.table.kwargs.get('mysql_engine')
        if ref_engine == 'MyISAM':
            # Don't try to create constraint when using MyISAM because it's
            # not supported.
            continue

        migrate.ForeignKeyConstraint(
            columns=[getattr(constraint_def['table'].c,
                             constraint_def['fk_column'])],
            refcolumns=[constraint_def['ref_column']]).create()


def find_repo(repo_name):
    """Return the absolute path to the named repository."""
    path = os.path.abspath(os.path.join(
        os.path.dirname(sql.__file__), repo_name))

    if not os.path.isdir(path):
        raise exception.MigrationNotProvided(sql.__name__, path)

    return path


def _sync_common_repo(version):
    abs_path = find_repo(LEGACY_REPO)
    init_version = get_init_version()
    with sql.session_for_write() as session:
        engine = session.get_bind()
        _assert_not_schema_downgrade(version=version)
        migration.db_sync(engine, abs_path, version=version,
                          init_version=init_version, sanity_check=False)


def _sync_repo(repo_name):
    abs_path = find_repo(repo_name)
    with sql.session_for_write() as session:
        engine = session.get_bind()
        # Register the repo with the version control API
        # If it already knows about the repo, it will throw
        # an exception that we can safely ignore
        try:
            migration.db_version_control(engine, abs_path)
        except (migration.exception.DBMigrationError,
                exceptions.DatabaseAlreadyControlledError):  # nosec
            pass
        init_version = get_init_version(abs_path=abs_path)
        migration.db_sync(engine, abs_path,
                          init_version=init_version, sanity_check=False)


def get_init_version(abs_path=None):
    """Get the initial version of a migrate repository.

    :param abs_path: Absolute path to migrate repository.
    :return:         initial version number or None, if DB is empty.
    """
    if abs_path is None:
        abs_path = find_repo(LEGACY_REPO)

    repo = migrate.versioning.repository.Repository(abs_path)

    # Sadly, Repository has a `latest` but not an `oldest`.
    # The value is a VerNum object which needs to be converted into an int.
    oldest = int(min(repo.versions.versions))

    if oldest < 1:
        return None

    # The initial version is one less
    return oldest - 1


def _assert_not_schema_downgrade(version=None):
    if version is not None:
        try:
            current_ver = int(six.text_type(get_db_version()))
            if int(version) < current_ver:
                raise migration.exception.DBMigrationError(
                    _("Unable to downgrade schema"))
        except exceptions.DatabaseNotControlledError:  # nosec
            # NOTE(morganfainberg): The database is not controlled, this action
            # cannot be a downgrade.
            pass


def offline_sync_database_to_version(version=None):
    """Perform and off-line sync of the database.

    Migrate the database up to the latest version, doing the equivalent of
    the cycle of --expand, --migrate and --contract, for when an offline
    upgrade is being performed.

    If a version is specified then only migrate the database up to that
    version. Downgrading is not supported. If version is specified, then only
    the main database migration is carried out - and the expand, migration and
    contract phases will NOT be run.

    """
    global USE_TRIGGERS

    # This flags let's us bypass trigger setup & teardown for non-rolling
    # upgrades. We set this as a global variable immediately before handing off
    # to sqlalchemy-migrate, because we can't pass arguments directly to
    # migrations that depend on it. We could also register this as a CONF
    # option, but the idea here is that we aren't exposing a new API.
    USE_TRIGGERS = False

    if version:
        _sync_common_repo(version)
    else:
        expand_schema()
        migrate_data()
        contract_schema()


def get_db_version(repo=LEGACY_REPO):
    with sql.session_for_read() as session:
        repo = find_repo(repo)
        return migration.db_version(
            session.get_bind(), repo, get_init_version(repo))


def validate_upgrade_order(repo_name, target_repo_version=None):
    """Validate the state of the migration repositories.

    This is run before allowing the db_sync command to execute. Ensure the
    upgrade step and version specified by the operator remains consistent with
    the upgrade process. I.e. expand's version is greater or equal to
    migrate's, migrate's version is greater or equal to contract's.

    :param repo_name: The name of the repository that the user is trying to
                      upgrade.
    :param target_repo_version: The version to upgrade the repo. Otherwise, the
                                version will be upgraded to the latest version
                                available.
    """
    # Initialize a dict to have each key assigned a repo with their value being
    # the repo that comes before.
    db_sync_order = {DATA_MIGRATION_REPO: EXPAND_REPO,
                     CONTRACT_REPO: DATA_MIGRATION_REPO}

    if repo_name == LEGACY_REPO:
        return
    # If expand is being run, we validate that Legacy repo is at the maximum
    # version before running the additional schema expansions.
    elif repo_name == EXPAND_REPO:
        abs_path = find_repo(LEGACY_REPO)
        repo = migrate.versioning.repository.Repository(abs_path)
        if int(repo.latest) != get_db_version():
            raise db_exception.DBMigrationError(
                'Your Legacy repo version is not up to date. Please refer to '
                'https://docs.openstack.org/keystone/latest/admin/'
                'identity-upgrading.html '
                'to see the proper steps for rolling upgrades.')
        return

    # find the latest version that the current command will upgrade to if there
    # wasn't a version specified for upgrade.
    if not target_repo_version:
        abs_path = find_repo(repo_name)
        repo = migrate.versioning.repository.Repository(abs_path)
        target_repo_version = int(repo.latest)

    # get current version of the command that runs before the current command.
    dependency_repo_version = get_db_version(repo=db_sync_order[repo_name])

    if dependency_repo_version < target_repo_version:
        raise db_exception.DBMigrationError(
            'You are attempting to upgrade %s ahead of %s. Please refer to '
            'https://docs.openstack.org/keystone/latest/admin/'
            'identity-upgrading.html '
            'to see the proper steps for rolling upgrades.' % (
                repo_name, db_sync_order[repo_name]))


def expand_schema():
    """Expand the database schema ahead of data migration.

    This is run manually by the keystone-manage command before the first
    keystone node is migrated to the latest release.

    """
    # Make sure all the legacy migrations are run before we run any new
    # expand migrations.
    _sync_common_repo(version=None)
    validate_upgrade_order(EXPAND_REPO)
    _sync_repo(repo_name=EXPAND_REPO)


def migrate_data():
    """Migrate data to match the new schema.

    This is run manually by the keystone-manage command once the keystone
    schema has been expanded for the new release.

    """
    validate_upgrade_order(DATA_MIGRATION_REPO)
    _sync_repo(repo_name=DATA_MIGRATION_REPO)


def contract_schema():
    """Contract the database.

    This is run manually by the keystone-manage command once the keystone
    nodes have been upgraded to the latest release and will remove any old
    tables/columns that are no longer required.

    """
    validate_upgrade_order(CONTRACT_REPO)
    _sync_repo(repo_name=CONTRACT_REPO)



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

import six

from keystone.common import provider_api
from keystone import exception


@six.add_metaclass(abc.ABCMeta)
class MappingDriverBase(provider_api.ProviderAPIMixin, object):
    """Interface description for an ID Mapping driver."""

    @abc.abstractmethod
    def get_public_id(self, local_entity):
        """Return the public ID for the given local entity.

        :param dict local_entity: Containing the entity domain, local ID and
                                  type ('user' or 'group').
        :returns: public ID, or None if no mapping is found.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_domain_mapping_list(self, domain_id, entity_type=None):
        """Return mappings for the domain.

        :param domain_id: Domain ID to get mappings for.
        :param entity_type: Optional entity_type to get mappings for.
        :type entity_type: String, one of mappings defined in
            keystone.identity.mapping_backends.mapping.EntityType
        :returns: list of mappings.
        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def get_id_mapping(self, public_id):
        """Return the local mapping.

        :param public_id: The public ID for the mapping required.
        :returns dict: Containing the entity domain, local ID and type. If no
                       mapping is found, it returns None.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def create_id_mapping(self, local_entity, public_id=None):
        """Create and store a mapping to a public_id.

        :param dict local_entity: Containing the entity domain, local ID and
                                  type ('user' or 'group').
        :param public_id: If specified, this will be the public ID.  If this
                          is not specified, a public ID will be generated.
        :returns: public ID

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def delete_id_mapping(self, public_id):
        """Delete an entry for the given public_id.

        :param public_id: The public ID for the mapping to be deleted.

        The method is silent if no mapping is found.

        """
        raise exception.NotImplemented()  # pragma: no cover

    @abc.abstractmethod
    def purge_mappings(self, purge_filter):
        """Purge selected identity mappings.

        :param dict purge_filter: Containing the attributes of the filter that
                                  defines which entries to purge. An empty
                                  filter means purge all mappings.

        """
        raise exception.NotImplemented()  # pragma: no cover



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import sqlalchemy

from keystone.common import sql
from keystone.endpoint_policy.backends import base
from keystone import exception


class PolicyAssociation(sql.ModelBase, sql.ModelDictMixin):
    __tablename__ = 'policy_association'
    attributes = ['policy_id', 'endpoint_id', 'region_id', 'service_id']
    # The id column is never exposed outside this module. It only exists to
    # provide a primary key, given that the real columns we would like to use
    # (endpoint_id, service_id, region_id) can be null
    id = sql.Column(sql.String(64), primary_key=True)
    policy_id = sql.Column(sql.String(64), nullable=False)
    endpoint_id = sql.Column(sql.String(64), nullable=True)
    service_id = sql.Column(sql.String(64), nullable=True)
    region_id = sql.Column(sql.String(64), nullable=True)
    __table_args__ = (sql.UniqueConstraint('endpoint_id', 'service_id',
                                           'region_id'),)

    def to_dict(self):
        """Return the model's attributes as a dictionary.

        We override the standard method in order to hide the id column,
        since this only exists to provide the table with a primary key.

        """
        d = {}
        for attr in self.__class__.attributes:
            d[attr] = getattr(self, attr)
        return d


class EndpointPolicy(base.EndpointPolicyDriverBase):

    def create_policy_association(self, policy_id, endpoint_id=None,
                                  service_id=None, region_id=None):
        with sql.session_for_write() as session:
            try:
                # See if there is already a row for this association, and if
                # so, update it with the new policy_id
                query = session.query(PolicyAssociation)
                query = query.filter_by(endpoint_id=endpoint_id)
                query = query.filter_by(service_id=service_id)
                query = query.filter_by(region_id=region_id)
                association = query.one()
                association.policy_id = policy_id
            except sql.NotFound:
                association = PolicyAssociation(id=uuid.uuid4().hex,
                                                policy_id=policy_id,
                                                endpoint_id=endpoint_id,
                                                service_id=service_id,
                                                region_id=region_id)
                session.add(association)

    def check_policy_association(self, policy_id, endpoint_id=None,
                                 service_id=None, region_id=None):
        sql_constraints = sqlalchemy.and_(
            PolicyAssociation.policy_id == policy_id,
            PolicyAssociation.endpoint_id == endpoint_id,
            PolicyAssociation.service_id == service_id,
            PolicyAssociation.region_id == region_id)

        # NOTE(henry-nash): Getting a single value to save object
        # management overhead.
        with sql.session_for_read() as session:
            if session.query(PolicyAssociation.id).filter(
                    sql_constraints).distinct().count() == 0:
                raise exception.PolicyAssociationNotFound()

    def delete_policy_association(self, policy_id, endpoint_id=None,
                                  service_id=None, region_id=None):
        with sql.session_for_write() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(policy_id=policy_id)
            query = query.filter_by(endpoint_id=endpoint_id)
            query = query.filter_by(service_id=service_id)
            query = query.filter_by(region_id=region_id)
            query.delete()

    def get_policy_association(self, endpoint_id=None,
                               service_id=None, region_id=None):
        sql_constraints = sqlalchemy.and_(
            PolicyAssociation.endpoint_id == endpoint_id,
            PolicyAssociation.service_id == service_id,
            PolicyAssociation.region_id == region_id)

        try:
            with sql.session_for_read() as session:
                policy_id = session.query(PolicyAssociation.policy_id).filter(
                    sql_constraints).distinct().one()
            return {'policy_id': policy_id}
        except sql.NotFound:
            raise exception.PolicyAssociationNotFound()

    def list_associations_for_policy(self, policy_id):
        with sql.session_for_read() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(policy_id=policy_id)
            return [ref.to_dict() for ref in query.all()]

    def delete_association_by_endpoint(self, endpoint_id):
        with sql.session_for_write() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(endpoint_id=endpoint_id)
            query.delete()

    def delete_association_by_service(self, service_id):
        with sql.session_for_write() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(service_id=service_id)
            query.delete()

    def delete_association_by_region(self, region_id):
        with sql.session_for_write() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(region_id=region_id)
            query.delete()

    def delete_association_by_policy(self, policy_id):
        with sql.session_for_write() as session:
            query = session.query(PolicyAssociation)
            query = query.filter_by(policy_id=policy_id)
            query.delete()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/groups

import flask
import flask_restful
import functools
from six.moves import http_client

from keystone.common import json_home
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common import validation
import keystone.conf
from keystone import exception
from keystone.identity import schema
from keystone import notifications
from keystone.server import flask as ks_flask


CONF = keystone.conf.CONF
ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs


def _build_group_target_enforcement():
    target = {}
    try:
        target['group'] = PROVIDERS.identity_api.get_group(
            flask.request.view_args.get('group_id')
        )
    except exception.NotFound:  # nosec
        # Defer existance in the event the group doesn't exist, we'll
        # check this later anyway.
        pass

    return target


class GroupsResource(ks_flask.ResourceBase):
    collection_key = 'groups'
    member_key = 'group'
    get_member_from_driver = PROVIDERS.deferred_provider_lookup(
        api='identity_api', method='get_group')

    def get(self, group_id=None):
        if group_id is not None:
            return self._get_group(group_id)
        return self._list_groups()

    def _get_group(self, group_id):
        """Get a group reference.

        GET/HEAD /groups/{group_id}
        """
        ENFORCER.enforce_call(
            action='identity:get_group',
            build_target=_build_group_target_enforcement
        )
        return self.wrap_member(PROVIDERS.identity_api.get_group(group_id))

    def _list_groups(self):
        """List groups.

        GET/HEAD /groups
        """
        filters = ['domain_id', 'name']
        target = None
        if self.oslo_context.domain_id:
            target = {'group': {'domain_id': self.oslo_context.domain_id}}
        ENFORCER.enforce_call(action='identity:list_groups', filters=filters,
                              target_attr=target)
        hints = self.build_driver_hints(filters)
        domain = self._get_domain_id_for_list_request()
        refs = PROVIDERS.identity_api.list_groups(domain_scope=domain,
                                                  hints=hints)
        if self.oslo_context.domain_id:
            filtered_refs = []
            for ref in refs:
                if ref['domain_id'] == target['group']['domain_id']:
                    filtered_refs.append(ref)
            refs = filtered_refs
        return self.wrap_collection(refs, hints=hints)

    def post(self):
        """Create group.

        POST /groups
        """
        group = self.request_body_json.get('group', {})
        target = {'group': group}
        ENFORCER.enforce_call(
            action='identity:create_group', target_attr=target
        )
        validation.lazy_validate(schema.group_create, group)
        group = self._normalize_dict(group)
        group = self._normalize_domain_id(group)
        ref = PROVIDERS.identity_api.create_group(
            group, initiator=self.audit_initiator)
        return self.wrap_member(ref), http_client.CREATED

    def patch(self, group_id):
        """Update group.

        PATCH /groups/{group_id}
        """
        ENFORCER.enforce_call(
            action='identity:update_group',
            build_target=_build_group_target_enforcement
        )
        group = self.request_body_json.get('group', {})
        validation.lazy_validate(schema.group_update, group)
        self._require_matching_id(group)
        ref = PROVIDERS.identity_api.update_group(
            group_id, group, initiator=self.audit_initiator)
        return self.wrap_member(ref)

    def delete(self, group_id):
        """Delete group.

        DELETE /groups/{group_id}
        """
        ENFORCER.enforce_call(action='identity:delete_group')
        PROVIDERS.identity_api.delete_group(
            group_id, initiator=self.audit_initiator)
        return None, http_client.NO_CONTENT


class GroupUsersResource(ks_flask.ResourceBase):
    def get(self, group_id):
        """Get list of users in group.

        GET/HEAD /groups/{group_id}/users
        """
        filters = ['domain_id', 'enabled', 'name', 'password_expires_at']
        target = None
        try:
            target = {'group': PROVIDERS.identity_api.get_group(group_id)}
        except exception.GroupNotFound:
            # NOTE(morgan): If we have an issue populating the group
            # data, leage target empty. This is the safest route and does not
            # leak data before enforcement happens.
            pass
        ENFORCER.enforce_call(action='identity:list_users_in_group',
                              target_attr=target, filters=filters)
        hints = ks_flask.ResourceBase.build_driver_hints(filters)
        refs = PROVIDERS.identity_api.list_users_in_group(
            group_id, hints=hints)
        if (self.oslo_context.domain_id):
            filtered_refs = []
            for ref in refs:
                if ref['domain_id'] == self.oslo_context.domain_id:
                    filtered_refs.append(ref)
            refs = filtered_refs
        return ks_flask.ResourceBase.wrap_collection(
            refs, hints=hints, collection_name='users')


class UserGroupCRUDResource(flask_restful.Resource):
    @staticmethod
    def _build_enforcement_target_attr(user_id, group_id):
        target = {}
        try:
            target['group'] = PROVIDERS.identity_api.get_group(group_id)
        except exception.GroupNotFound:
            # Don't populate group data if group is not found.
            pass

        try:
            target['user'] = PROVIDERS.identity_api.get_user(user_id)
        except exception.UserNotFound:
            # Don't populate user data if user is not found
            pass

        return target

    def get(self, group_id, user_id):
        """Check if a user is in a group.

        GET/HEAD /groups/{group_id}/users/{user_id}
        """
        ENFORCER.enforce_call(
            action='identity:check_user_in_group',
            build_target=functools.partial(self._build_enforcement_target_attr,
                                           user_id, group_id))
        PROVIDERS.identity_api.check_user_in_group(user_id, group_id)
        return None, http_client.NO_CONTENT

    def put(self, group_id, user_id):
        """Add user to group.

        PUT /groups/{group_id}/users/{user_id}
        """
        ENFORCER.enforce_call(
            action='identity:add_user_to_group',
            build_target=functools.partial(self._build_enforcement_target_attr,
                                           user_id, group_id))
        PROVIDERS.identity_api.add_user_to_group(
            user_id, group_id, initiator=notifications.build_audit_initiator())
        return None, http_client.NO_CONTENT

    def delete(self, group_id, user_id):
        """Remove user from group.

        DELETE /groups/{group_id}/users/{user_id}
        """
        ENFORCER.enforce_call(
            action='identity:remove_user_from_group',
            build_target=functools.partial(self._build_enforcement_target_attr,
                                           user_id, group_id))
        PROVIDERS.identity_api.remove_user_from_group(
            user_id, group_id, initiator=notifications.build_audit_initiator())
        return None, http_client.NO_CONTENT


class GroupAPI(ks_flask.APIBase):
    _name = 'groups'
    _import_name = __name__
    resources = [GroupsResource]
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=GroupUsersResource,
            url='/groups/<string:group_id>/users',
            resource_kwargs={},
            rel='group_users',
            path_vars={'group_id': json_home.Parameters.GROUP_ID}),
        ks_flask.construct_resource_map(
            resource=UserGroupCRUDResource,
            url='/groups/<string:group_id>/users/<string:user_id>',
            resource_kwargs={},
            rel='group_user',
            path_vars={
                'group_id': json_home.Parameters.GROUP_ID,
                'user_id': json_home.Parameters.USER_ID})
    ]


APIs = (GroupAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy as sql

from keystone.common.sql import upgrades

# define the local_user triggers for insert and update
MYSQL_LOCAL_USER_INSERT_TRIGGER = """
CREATE TRIGGER local_user_after_insert_trigger
AFTER INSERT
    ON local_user FOR EACH ROW
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id IS NULL;
END;
"""

MYSQL_LOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER local_user_after_update_trigger
AFTER UPDATE
    ON local_user FOR EACH ROW
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id <> NEW.domain_id;
END;
"""

SQLITE_LOCAL_USER_INSERT_TRIGGER = """
CREATE TRIGGER local_user_after_insert_trigger
AFTER INSERT
    ON local_user
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id IS NULL;
END;
"""

SQLITE_LOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER local_user_after_update_trigger
AFTER UPDATE
    ON local_user
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id <> NEW.domain_id;
END;
"""

POSTGRESQL_LOCAL_USER_INSERT_TRIGGER = """
CREATE OR REPLACE FUNCTION update_user_domain_id()
    RETURNS trigger AS
$BODY$
BEGIN
    UPDATE "user" SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id;
    RETURN NULL;
END
$BODY$ LANGUAGE plpgsql;

CREATE TRIGGER local_user_after_insert_trigger AFTER INSERT ON local_user
FOR EACH ROW
EXECUTE PROCEDURE update_user_domain_id();
"""

POSTGRESQL_LOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER local_user_after_update_trigger AFTER UPDATE ON local_user
FOR EACH ROW
EXECUTE PROCEDURE update_user_domain_id();
"""

MYSQL_NONLOCAL_USER_INSERT_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_insert_trigger
AFTER INSERT
    ON nonlocal_user FOR EACH ROW
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id IS NULL;
END;
"""

# define the nonlocal_user triggers for insert and update
MYSQL_NONLOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_update_trigger
AFTER UPDATE
    ON nonlocal_user FOR EACH ROW
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id <> NEW.domain_id;
END;
"""

SQLITE_NONLOCAL_USER_INSERT_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_insert_trigger
AFTER INSERT
    ON nonlocal_user
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id IS NULL;
END;
"""

SQLITE_NONLOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_update_trigger
AFTER UPDATE
    ON nonlocal_user
BEGIN
    UPDATE user SET domain_id = NEW.domain_id
        WHERE id = NEW.user_id and domain_id <> NEW.domain_id;
END;
"""

POSTGRESQL_NONLOCAL_USER_INSERT_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_insert_trigger AFTER INSERT ON nonlocal_user
FOR EACH ROW
EXECUTE PROCEDURE update_user_domain_id();
"""

POSTGRESQL_NONLOCAL_USER_UPDATE_TRIGGER = """
CREATE TRIGGER nonlocal_user_after_update_trigger AFTER UPDATE ON nonlocal_user
FOR EACH ROW
EXECUTE PROCEDURE update_user_domain_id();
"""


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    user = sql.Table('user', meta, autoload=True)
    project = sql.Table('project', meta, autoload=True)
    domain_id = sql.Column('domain_id', sql.String(64),
                           sql.ForeignKey(project.c.id), nullable=True)
    user.create_column(domain_id)

    if upgrades.USE_TRIGGERS:
        if migrate_engine.name == 'postgresql':
            local_user_insert_trigger = POSTGRESQL_LOCAL_USER_INSERT_TRIGGER
            local_user_update_trigger = POSTGRESQL_LOCAL_USER_UPDATE_TRIGGER
            nonlocal_user_insert_trigger = (
                POSTGRESQL_NONLOCAL_USER_INSERT_TRIGGER)
            nonlocal_user_update_trigger = (
                POSTGRESQL_NONLOCAL_USER_UPDATE_TRIGGER)
        elif migrate_engine.name == 'sqlite':
            local_user_insert_trigger = SQLITE_LOCAL_USER_INSERT_TRIGGER
            local_user_update_trigger = SQLITE_LOCAL_USER_UPDATE_TRIGGER
            nonlocal_user_insert_trigger = SQLITE_NONLOCAL_USER_INSERT_TRIGGER
            nonlocal_user_update_trigger = SQLITE_NONLOCAL_USER_UPDATE_TRIGGER
        else:
            local_user_insert_trigger = MYSQL_LOCAL_USER_INSERT_TRIGGER
            local_user_update_trigger = MYSQL_LOCAL_USER_UPDATE_TRIGGER
            nonlocal_user_insert_trigger = MYSQL_NONLOCAL_USER_INSERT_TRIGGER
            nonlocal_user_update_trigger = MYSQL_NONLOCAL_USER_UPDATE_TRIGGER
        migrate_engine.execute(local_user_insert_trigger)
        migrate_engine.execute(local_user_update_trigger)
        migrate_engine.execute(nonlocal_user_insert_trigger)
        migrate_engine.execute(nonlocal_user_update_trigger)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import provider_api
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import default_fixtures
from keystone.tests.unit.ksfixtures import database
from keystone.tests.unit import mapping_fixtures

PROVIDERS = provider_api.ProviderAPIs


class TestFederationProtocol(unit.TestCase):

    def setUp(self):
        super(TestFederationProtocol, self).setUp()
        self.useFixture(database.Database())
        self.load_backends()
        PROVIDERS.resource_api.create_domain(
            default_fixtures.ROOT_DOMAIN['id'], default_fixtures.ROOT_DOMAIN)
        self.idp = {
            'id': uuid.uuid4().hex,
            'enabled': True,
            'description': uuid.uuid4().hex
        }
        PROVIDERS.federation_api.create_idp(self.idp['id'], self.idp)
        self.mapping = mapping_fixtures.MAPPING_EPHEMERAL_USER
        self.mapping['id'] = uuid.uuid4().hex
        PROVIDERS.federation_api.create_mapping(
            self.mapping['id'], self.mapping
        )

    def test_create_protocol(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': self.mapping['id']
        }
        protocol_ret = PROVIDERS.federation_api.create_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['id'], protocol_ret['id'])

    def test_create_protocol_with_invalid_mapping_id(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': uuid.uuid4().hex
        }
        self.assertRaises(exception.ValidationError,
                          PROVIDERS.federation_api.create_protocol,
                          self.idp['id'],
                          protocol['id'],
                          protocol)

    def test_create_protocol_with_remote_id_attribute(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': self.mapping['id'],
            'remote_id_attribute': uuid.uuid4().hex
        }
        protocol_ret = PROVIDERS.federation_api.create_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['remote_id_attribute'],
                         protocol_ret['remote_id_attribute'])

    def test_update_protocol(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': self.mapping['id']
        }
        protocol_ret = PROVIDERS.federation_api.create_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['id'], protocol_ret['id'])
        new_mapping = mapping_fixtures.MAPPING_EPHEMERAL_USER
        new_mapping['id'] = uuid.uuid4().hex
        PROVIDERS.federation_api.create_mapping(new_mapping['id'], new_mapping)
        protocol['mapping_id'] = new_mapping['id']
        protocol_ret = PROVIDERS.federation_api.update_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['id'], protocol_ret['id'])
        self.assertEqual(new_mapping['id'], protocol_ret['mapping_id'])

    def test_update_protocol_with_invalid_mapping_id(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': self.mapping['id']
        }
        protocol_ret = PROVIDERS.federation_api.create_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['id'], protocol_ret['id'])
        protocol['mapping_id'] = uuid.uuid4().hex
        self.assertRaises(exception.ValidationError,
                          PROVIDERS.federation_api.update_protocol,
                          self.idp['id'],
                          protocol['id'],
                          protocol)

    def test_update_protocol_with_remote_id_attribute(self):
        protocol = {
            'id': uuid.uuid4().hex,
            'mapping_id': self.mapping['id']
        }
        protocol_ret = PROVIDERS.federation_api.create_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        new_remote_id_attribute = uuid.uuid4().hex
        protocol['remote_id_attribute'] = new_remote_id_attribute
        protocol_ret = PROVIDERS.federation_api.update_protocol(
            self.idp['id'], protocol['id'], protocol
        )
        self.assertEqual(protocol['remote_id_attribute'],
                         protocol_ret['remote_id_attribute'])



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneclient.contrib.ec2 import utils as ec2_utils
from six.moves import http_client

from keystone.common import provider_api
from keystone.tests import unit
from keystone.tests.unit import test_v3


PROVIDERS = provider_api.ProviderAPIs


class EC2ContribCoreV3(test_v3.RestfulTestCase):
    def setUp(self):
        super(EC2ContribCoreV3, self).setUp()

        self.cred_blob, self.credential = unit.new_ec2_credential(
            self.user['id'], self.project_id)
        PROVIDERS.credential_api.create_credential(
            self.credential['id'], self.credential)

    def test_valid_authentication_response_with_proper_secret(self):
        signer = ec2_utils.Ec2Signer(self.cred_blob['secret'])
        credentials = {
            'access': self.cred_blob['access'],
            'secret': self.cred_blob['secret'],
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'SignatureVersion': '2',
                'Action': 'Test',
                'Timestamp': '2007-01-31T23:59:59Z'
            },
        }
        credentials['signature'] = signer.generate(credentials)
        resp = self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http_client.OK)
        self.assertValidProjectScopedTokenResponse(resp, self.user)

    def test_authenticate_with_empty_body_returns_bad_request(self):
        self.post(
            '/ec2tokens',
            body={},
            expected_status=http_client.BAD_REQUEST)

    def test_authenticate_without_json_request_returns_bad_request(self):
        self.post(
            '/ec2tokens',
            body='not json',
            expected_status=http_client.BAD_REQUEST)

    def test_authenticate_without_request_body_returns_bad_request(self):
        self.post(
            '/ec2tokens',
            expected_status=http_client.BAD_REQUEST)

    def test_authenticate_without_proper_secret_returns_unauthorized(self):
        signer = ec2_utils.Ec2Signer('totally not the secret')
        credentials = {
            'access': self.cred_blob['access'],
            'secret': 'totally not the secret',
            'host': 'localhost',
            'verb': 'GET',
            'path': '/',
            'params': {
                'SignatureVersion': '2',
                'Action': 'Test',
                'Timestamp': '2007-01-31T23:59:59Z'
            },
        }
        credentials['signature'] = signer.generate(credentials)
        self.post(
            '/ec2tokens',
            body={'credentials': credentials},
            expected_status=http_client.UNAUTHORIZED)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import resource_options
from keystone.common.resource_options import options as ro_opt


ROLE_OPTIONS_REGISTRY = resource_options.ResourceOptionRegistry('ROLE')


# NOTE(morgan): wrap this in a function for testing purposes.
# This is called on import by design.
def register_role_options():
    for opt in [
        ro_opt.IMMUTABLE_OPT,
    ]:
        ROLE_OPTIONS_REGISTRY.register_option(opt)


register_role_options()



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_utils import timeutils
from six.moves import urllib

from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.models import token_model
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import database
from keystone import token
from keystone.token import provider


CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs

FUTURE_DELTA = datetime.timedelta(seconds=CONF.token.expiration)
CURRENT_DATE = timeutils.utcnow()


class TestTokenProvider(unit.TestCase):
    def setUp(self):
        super(TestTokenProvider, self).setUp()
        self.useFixture(database.Database())
        self.useFixture(
            ksfixtures.KeyRepository(
                self.config_fixture,
                'fernet_tokens',
                CONF.fernet_tokens.max_active_keys
            )
        )
        self.load_backends()

    def test_strings_are_url_safe(self):
        s = provider.random_urlsafe_str()
        self.assertEqual(s, urllib.parse.quote_plus(s))

    def test_unsupported_token_provider(self):
        self.config_fixture.config(group='token',
                                   provider='MyProvider')
        self.assertRaises(ImportError,
                          token.provider.Manager)

    def test_provider_token_expiration_validation(self):
        token = token_model.TokenModel()
        token.issued_at = "2013-05-21T00:02:43.941473Z"
        token.expires_at = utils.isotime(CURRENT_DATE)

        self.assertRaises(exception.TokenNotFound,
                          PROVIDERS.token_provider_api._is_valid_token,
                          token)

        token = token_model.TokenModel()
        token.issued_at = "2013-05-21T00:02:43.941473Z"
        token.expires_at = utils.isotime(timeutils.utcnow() + FUTURE_DELTA)
        self.assertIsNone(PROVIDERS.token_provider_api._is_valid_token(token))

    def test_validate_v3_token_with_no_token_raises_token_not_found(self):
        self.assertRaises(
            exception.TokenNotFound,
            PROVIDERS.token_provider_api.validate_token,
            None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

import sqlalchemy as sql
from sqlalchemy.orm import sessionmaker

from keystone.resource.backends import base


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine
    maker = sessionmaker(bind=migrate_engine)
    session = maker()

    idp_table = sql.Table('identity_provider', meta, autoload=True)

    for idp_row in idp_table.select().execute():
        domain_id = _create_federated_domain(meta, session, idp_row['id'])
        # update idp with the new federated domain_id
        values = {'domain_id': domain_id}
        stmt = idp_table.update().where(
            idp_table.c.id == idp_row['id']).values(values)
        stmt.execute()


def _create_federated_domain(meta, session, idp_id):
    domain_id = uuid.uuid4().hex
    desc = 'Auto generated federated domain for Identity Provider: ' + idp_id
    federated_domain = {
        'id': domain_id,
        'name': domain_id,
        'enabled': True,
        'description': desc,
        'domain_id': base.NULL_DOMAIN_ID,
        'is_domain': True,
        'parent_id': None,
        'extra': '{}'
    }
    project_table = sql.Table('project', meta, autoload=True)
    new_row = project_table.insert().values(**federated_domain)
    session.execute(new_row)
    session.commit()
    return domain_id



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import os.path

from oslo_log import log

from keystone.catalog.backends import base
from keystone.common import utils
import keystone.conf
from keystone import exception


LOG = log.getLogger(__name__)

CONF = keystone.conf.CONF


def parse_templates(template_lines):
    o = {}
    for line in template_lines:
        if ' = ' not in line:
            continue

        k, v = line.strip().split(' = ')
        if not k.startswith('catalog.'):
            continue

        parts = k.split('.')

        region = parts[1]
        # NOTE(termie): object-store insists on having a dash
        service = parts[2].replace('_', '-')
        key = parts[3]

        region_ref = o.get(region, {})
        service_ref = region_ref.get(service, {})
        service_ref[key] = v

        region_ref[service] = service_ref
        o[region] = region_ref

    return o


class Catalog(base.CatalogDriverBase):
    """A backend that generates endpoints for the Catalog based on templates.

    It is usually configured via config entries that look like:

      catalog.$REGION.$SERVICE.$key = $value

    and is stored in a similar looking hierarchy. Where a value can contain
    values to be interpolated by standard python string interpolation that look
    like (the % is replaced by a $):

      http://localhost:$(public_port)s/

    When expanding the template it will pass in a dict made up of the conf
    instance plus a few additional key-values, notably project_id and user_id.

    It does not care what the keys and values are but it is worth noting that
    keystone_compat will expect certain keys to be there so that it can munge
    them into the output format keystone expects. These keys are:

      name - the name of the service, most likely repeated for all services of
             the same type, across regions.

      adminURL - the url of the admin endpoint

      publicURL - the url of the public endpoint

      internalURL - the url of the internal endpoint

    """

    def __init__(self, templates=None):
        super(Catalog, self).__init__()
        if templates:
            self.templates = templates
        else:
            template_file = CONF.catalog.template_file
            if not os.path.exists(template_file):
                template_file = CONF.find_file(template_file)
            self._load_templates(template_file)

    def _load_templates(self, template_file):
        try:
            with open(template_file) as f:
                self.templates = parse_templates(f)
        except IOError:
            LOG.critical('Unable to open template file %s', template_file)
            raise

    # region crud

    def create_region(self, region_ref):
        raise exception.NotImplemented()

    def list_regions(self, hints):
        return [{'id': region_id, 'description': '', 'parent_region_id': ''}
                for region_id in self.templates]

    def get_region(self, region_id):
        if region_id in self.templates:
            return {'id': region_id, 'description': '', 'parent_region_id': ''}
        raise exception.RegionNotFound(region_id=region_id)

    def update_region(self, region_id, region_ref):
        raise exception.NotImplemented()

    def delete_region(self, region_id):
        raise exception.NotImplemented()

    # service crud

    def create_service(self, service_id, service_ref):
        raise exception.NotImplemented()

    def _list_services(self, hints):
        for region_ref in self.templates.values():
            for service_type, service_ref in region_ref.items():
                yield {
                    'id': service_type,
                    'enabled': True,
                    'name': service_ref.get('name', ''),
                    'description': service_ref.get('description', ''),
                    'type': service_type,
                }

    def list_services(self, hints):
        return list(self._list_services(hints=None))

    def get_service(self, service_id):
        for service in self._list_services(hints=None):
            if service['id'] == service_id:
                return service
        raise exception.ServiceNotFound(service_id=service_id)

    def update_service(self, service_id, service_ref):
        raise exception.NotImplemented()

    def delete_service(self, service_id):
        raise exception.NotImplemented()

    # endpoint crud

    def create_endpoint(self, endpoint_id, endpoint_ref):
        raise exception.NotImplemented()

    def _list_endpoints(self):
        for region_id, region_ref in self.templates.items():
            for service_type, service_ref in region_ref.items():
                for key in service_ref:
                    if key.endswith('URL'):
                        interface = key[:-3]
                        endpoint_id = ('%s-%s-%s' %
                                       (region_id, service_type, interface))
                        yield {
                            'id': endpoint_id,
                            'service_id': service_type,
                            'interface': interface,
                            'url': service_ref[key],
                            'legacy_endpoint_id': None,
                            'region_id': region_id,
                            'enabled': True,
                        }

    def list_endpoints(self, hints):
        return list(self._list_endpoints())

    def get_endpoint(self, endpoint_id):
        for endpoint in self._list_endpoints():
            if endpoint['id'] == endpoint_id:
                return endpoint
        raise exception.EndpointNotFound(endpoint_id=endpoint_id)

    def update_endpoint(self, endpoint_id, endpoint_ref):
        raise exception.NotImplemented()

    def delete_endpoint(self, endpoint_id):
        raise exception.NotImplemented()

    def get_catalog(self, user_id, project_id):
        """Retrieve and format the V2 service catalog.

        :param user_id: The id of the user who has been authenticated for
            creating service catalog.
        :param project_id: The id of the project. 'project_id' will be None in
            the case this being called to create a catalog to go in a domain
            scoped token. In this case, any endpoint that requires a project_id
            as part of their URL will be skipped.

        :returns: A nested dict representing the service catalog or an
                  empty dict.

        """
        substitutions = dict(
            itertools.chain(CONF.items(), CONF.eventlet_server.items()))
        substitutions.update({'user_id': user_id})
        silent_keyerror_failures = []
        if project_id:
            substitutions.update({
                'tenant_id': project_id,
                'project_id': project_id,
            })
        else:
            silent_keyerror_failures = ['tenant_id', 'project_id', ]

        catalog = {}
        # TODO(davechen): If there is service with no endpoints, we should
        # skip the service instead of keeping it in the catalog.
        # see bug #1436704.
        for region, region_ref in self.templates.items():
            catalog[region] = {}
            for service, service_ref in region_ref.items():
                service_data = {}
                try:
                    for k, v in service_ref.items():
                        formatted_value = utils.format_url(
                            v, substitutions,
                            silent_keyerror_failures=silent_keyerror_failures)
                        if formatted_value:
                            service_data[k] = formatted_value
                except exception.MalformedEndpoint:  # nosec(tkelsey)
                    continue  # this failure is already logged in format_url()
                catalog[region][service] = service_data

        return catalog

    def get_v3_catalog(self, user_id, project_id):
        """Retrieve and format the current V3 service catalog.

        This implementation builds the V3 catalog from the V2 catalog.

        :param user_id: The id of the user who has been authenticated for
            creating service catalog.
        :param project_id: The id of the project. 'project_id' will be None in
            the case this being called to create a catalog to go in a domain
            scoped token. In this case, any endpoint that requires a project_id
            as part of their URL will be skipped.

        :returns: A list representing the service catalog or an empty list

        """
        v2_catalog = self.get_catalog(user_id, project_id)
        v3_catalog = {}

        for region_name, region in v2_catalog.items():
            for service_type, service in region.items():
                if service_type not in v3_catalog:
                    v3_catalog[service_type] = {
                        'type': service_type,
                        'endpoints': []
                    }

                for attr, value in service.items():
                    # Attributes that end in URL are interfaces. In the V2
                    # catalog, these are internalURL, publicURL, and adminURL.
                    # For example, <region_name>.publicURL=<URL> in the V2
                    # catalog becomes the V3 interface for the service:
                    # { 'interface': 'public', 'url': '<URL>', 'region':
                    #   'region: '<region_name>' }
                    if attr.endswith('URL'):
                        v3_interface = attr[:-len('URL')]
                        v3_catalog[service_type]['endpoints'].append({
                            'interface': v3_interface,
                            'region': region_name,
                            'url': value,
                        })
                        continue

                    # Other attributes are copied to the service.
                    v3_catalog[service_type][attr] = value

        return list(v3_catalog.values())

    def add_endpoint_to_project(self, endpoint_id, project_id):
        raise exception.NotImplemented()

    def remove_endpoint_from_project(self, endpoint_id, project_id):
        raise exception.NotImplemented()

    def check_endpoint_in_project(self, endpoint_id, project_id):
        raise exception.NotImplemented()

    def list_endpoints_for_project(self, project_id):
        raise exception.NotImplemented()

    def list_projects_for_endpoint(self, endpoint_id):
        raise exception.NotImplemented()

    def delete_association_by_endpoint(self, endpoint_id):
        raise exception.NotImplemented()

    def delete_association_by_project(self, project_id):
        raise exception.NotImplemented()

    def create_endpoint_group(self, endpoint_group):
        raise exception.NotImplemented()

    def get_endpoint_group(self, endpoint_group_id):
        raise exception.NotImplemented()

    def update_endpoint_group(self, endpoint_group_id, endpoint_group):
        raise exception.NotImplemented()

    def delete_endpoint_group(self, endpoint_group_id):
        raise exception.NotImplemented()

    def add_endpoint_group_to_project(self, endpoint_group_id, project_id):
        raise exception.NotImplemented()

    def get_endpoint_group_in_project(self, endpoint_group_id, project_id):
        raise exception.NotImplemented()

    def list_endpoint_groups(self, hints):
        raise exception.NotImplemented()

    def list_endpoint_groups_for_project(self, project_id):
        raise exception.NotImplemented()

    def list_projects_associated_with_endpoint_group(self, endpoint_group_id):
        raise exception.NotImplemented()

    def remove_endpoint_group_from_project(self, endpoint_group_id,
                                           project_id):
        raise exception.NotImplemented()

    def delete_endpoint_group_association_by_project(self, project_id):
        raise exception.NotImplemented()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import migrate
import sqlalchemy as sql


MYSQL_CREATE_ID_PRIMARY_KEY_COLUMN = """
ALTER TABLE `%s` ADD `internal_id` INT NOT NULL AUTO_INCREMENT PRIMARY KEY;
"""

POSTGRESQL_CREATE_ID_PRIMARY_KEY_COLUMN = """
ALTER TABLE "%s" ADD COLUMN "internal_id" SERIAL PRIMARY KEY;
"""


def upgrade(migrate_engine):

    # For both registered_limit and limit tables in MySQL and PostgreSQL:
    #
    # 1. drop the primary key on `id` column.
    # 2. create a auto increment `internal_id` column with primary key.
    # 3. add unique constraint on `id` column.
    #
    # But SQLite doesn't support add primary key to a existed table, so for
    # SQLite, we'll follow the steps, take the registered_limit as an example:
    #
    # 1. Add a new table `registered_limit_new` which contains `internal_id`
    #    column.
    # 2. migrate the data from `registered_limit` to `registered_limit_new`
    # 3. drop the `registered_limit`, rename `registered_limit_new` to
    #    `registered_limit`.

    meta = sql.MetaData()
    meta.bind = migrate_engine
    registered_limit_table = sql.Table('registered_limit', meta, autoload=True)
    limit_table = sql.Table('limit', meta, autoload=True)

    if migrate_engine.name != 'sqlite':
        pk = migrate.PrimaryKeyConstraint('id', table=registered_limit_table)
        pk.drop()
        if migrate_engine.name == 'mysql':
            migrate_engine.execute(
                MYSQL_CREATE_ID_PRIMARY_KEY_COLUMN % 'registered_limit')
        else:
            migrate_engine.execute(
                POSTGRESQL_CREATE_ID_PRIMARY_KEY_COLUMN % 'registered_limit')
        unique_constraint = migrate.UniqueConstraint(
            'id', table=registered_limit_table)
        unique_constraint.create()

        pk = migrate.PrimaryKeyConstraint('id', table=limit_table)
        pk.drop()
        if migrate_engine.name == 'mysql':
            migrate_engine.execute(
                MYSQL_CREATE_ID_PRIMARY_KEY_COLUMN % 'limit')
        else:
            migrate_engine.execute(
                POSTGRESQL_CREATE_ID_PRIMARY_KEY_COLUMN % 'limit')
        unique_constraint = migrate.UniqueConstraint('id', table=limit_table)
        unique_constraint.create()
    else:
        # SQLite case
        registered_limit_table_new = sql.Table(
            'registered_limit_new',
            meta,
            sql.Column('internal_id', sql.Integer, primary_key=True),
            sql.Column('id', sql.String(length=64), unique=True),
            sql.Column('service_id',
                       sql.String(64)),
            sql.Column('region_id',
                       sql.String(64),
                       nullable=True),
            sql.Column('resource_name', sql.String(255)),
            sql.Column('default_limit', sql.Integer, nullable=False),
            sql.Column('description', sql.Text),
            mysql_engine='InnoDB',
            mysql_charset='utf8')
        registered_limit_table_new.create(migrate_engine, checkfirst=True)

        limit_table_new = sql.Table(
            'limit_new',
            meta,
            sql.Column('internal_id', sql.Integer, primary_key=True),
            sql.Column('id', sql.String(length=64), unique=True),
            sql.Column('project_id', sql.String(64)),
            sql.Column('service_id', sql.String(64)),
            sql.Column('region_id', sql.String(64), nullable=True),
            sql.Column('resource_name', sql.String(255)),
            sql.Column('resource_limit', sql.Integer, nullable=False),
            sql.Column('description', sql.Text),
            mysql_engine='InnoDB',
            mysql_charset='utf8')
        limit_table_new.create(migrate_engine, checkfirst=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common.sql import upgrades

import sqlalchemy as sql


def upgrade(migrate_engine):
    meta = sql.MetaData()
    meta.bind = migrate_engine

    credential_table = sql.Table('credential', meta, autoload=True)
    credential_table.c.blob.drop()

    if upgrades.USE_TRIGGERS:
        if migrate_engine.name == 'postgresql':
            drop_credential_update_trigger = (
                'DROP TRIGGER credential_update_read_only on credential;'
            )
            drop_credential_insert_trigger = (
                'DROP TRIGGER credential_insert_read_only on credential;'
            )
        elif migrate_engine.name == 'mysql':
            drop_credential_update_trigger = (
                'DROP TRIGGER credential_update_read_only;'
            )
            drop_credential_insert_trigger = (
                'DROP TRIGGER credential_insert_read_only;'
            )
        else:
            # NOTE(lbragstad, henry-nash): Apparently sqlalchemy and sqlite
            # behave weird when using triggers, which is why we use the `IF
            # EXISTS` conditional here.  I think what is happening is that the
            # credential_table.c.blob.drop() causes sqlalchemy to create a new
            # credential table - but it doesn't copy the triggers over, which
            # causes the DROP TRIGGER statement to fail without `IF EXISTS`
            # because the trigger doesn't exist in the new table(?!).
            drop_credential_update_trigger = (
                'DROP TRIGGER IF EXISTS credential_update_read_only;'
            )
            drop_credential_insert_trigger = (
                'DROP TRIGGER IF EXISTS credential_insert_read_only;'
            )
        migrate_engine.execute(drop_credential_update_trigger)
        migrate_engine.execute(drop_credential_insert_trigger)

    # NOTE(lbragstad): We close these so that they are not nullable because
    # Newton code (and anything after) would always populate these values.
    credential_table.c.encrypted_blob.alter(nullable=False)
    credential_table.c.key_hash.alter(nullable=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone import exception


class DriverTestCase(object):
    """Test cases to validate the endpoint policy driver behavior."""

    @property
    def driver(self):
        raise exception.NotImplemented()

    def create_association(self, **kwargs):
        association = {'policy_id': uuid.uuid4().hex,
                       'endpoint_id': None,
                       'service_id': None,
                       'region_id': None}
        association.update(kwargs)
        self.driver.create_policy_association(**association)
        return association

    def test_create_policy_association(self):
        association = self.create_association(endpoint_id=uuid.uuid4().hex)
        self.driver.check_policy_association(**association)

        association = self.create_association(service_id=uuid.uuid4().hex,
                                              region_id=uuid.uuid4().hex)
        self.driver.check_policy_association(**association)

        association = self.create_association(service_id=uuid.uuid4().hex)
        self.driver.check_policy_association(**association)

    def test_recreate_policy_association(self):
        # Creating a policy association to a target that already has a policy
        # associated to it will cause the original policy to be overridden
        original_association = self.create_association(
            service_id=uuid.uuid4().hex)
        override_association = original_association.copy()
        override_association['policy_id'] = uuid.uuid4().hex

        self.driver.create_policy_association(**override_association)

        self.driver.check_policy_association(**override_association)
        self.assertRaises(exception.PolicyAssociationNotFound,
                          self.driver.check_policy_association,
                          **original_association)

    def test_check_policy_association(self):
        association = self.create_association(service_id=uuid.uuid4().hex,
                                              region_id=uuid.uuid4().hex)
        self.driver.check_policy_association(**association)

        # An association is uniquely identified by its target. Omitting any
        # attribute (region_id in this case) will result in a different check
        association.pop('region_id')

        self.assertRaises(exception.PolicyAssociationNotFound,
                          self.driver.check_policy_association,
                          **association)

    def test_delete_policy_association(self):
        association = self.create_association(endpoint_id=uuid.uuid4().hex)
        self.driver.delete_policy_association(**association)

        self.assertRaises(exception.PolicyAssociationNotFound,
                          self.driver.check_policy_association,
                          **association)

    def test_get_policy_association(self):
        association = self.create_association(service_id=uuid.uuid4().hex)

        # Extract the policy_id from the association and query it by the target
        policy_id = association.pop('policy_id')

        association_ref = self.driver.get_policy_association(**association)
        self.assertEqual({'policy_id': (policy_id,)}, association_ref)

    def test_list_associations_for_policy(self):
        policy_id = uuid.uuid4().hex
        first = self.create_association(endpoint_id=uuid.uuid4().hex,
                                        policy_id=policy_id)
        second = self.create_association(service_id=uuid.uuid4().hex,
                                         policy_id=policy_id)

        associations = self.driver.list_associations_for_policy(policy_id)
        self.assertItemsEqual([first, second], associations)

    def test_delete_association_by_endpoint(self):
        endpoint_id = uuid.uuid4().hex
        associations = [self.create_association(endpoint_id=endpoint_id),
                        self.create_association(endpoint_id=endpoint_id)]

        self.driver.delete_association_by_endpoint(endpoint_id)

        for association in associations:
            self.assertRaises(exception.PolicyAssociationNotFound,
                              self.driver.check_policy_association,
                              **association)

    def test_delete_association_by_service(self):
        service_id = uuid.uuid4().hex
        associations = [self.create_association(service_id=service_id),
                        self.create_association(service_id=service_id)]

        self.driver.delete_association_by_service(service_id)

        for association in associations:
            self.assertRaises(exception.PolicyAssociationNotFound,
                              self.driver.check_policy_association,
                              **association)

    def test_delete_association_by_region(self):
        region_id = uuid.uuid4().hex
        first = self.create_association(service_id=uuid.uuid4().hex,
                                        region_id=region_id)
        second = self.create_association(service_id=uuid.uuid4().hex,
                                         region_id=region_id)

        self.driver.delete_association_by_region(region_id)

        for association in [first, second]:
            self.assertRaises(exception.PolicyAssociationNotFound,
                              self.driver.check_policy_association,
                              **association)

    def test_delete_association_by_policy(self):
        policy_id = uuid.uuid4().hex
        first = self.create_association(endpoint_id=uuid.uuid4().hex,
                                        policy_id=policy_id)
        second = self.create_association(service_id=uuid.uuid4().hex,
                                         policy_id=policy_id)

        self.driver.delete_association_by_policy(policy_id)

        for association in [first, second]:
            self.assertRaises(exception.PolicyAssociationNotFound,
                              self.driver.check_policy_association,
                              **association)
